/***
 * $Source: /usr/home/ldl/langs/fm/RCS/compile.c,v $
 * $Revision: 0.19980909 $ $Date: 1998/09/09 20:46:27 $
 *
 * FM - Femto-M[UMPS].
 * Compile program into internal structures.
 */

#include "fm.h"
#include "opcode.h"
#include <ctype.h>

/**
 *	Local functions.
 */
static
Command command		PROTO_ARGS((char **));
static
char   *copy_non_white	PROTO_ARGS((char **, char *));
static
char   *copy_non_white_upto PROTO_ARGS((char **, char *, char *));
static
void	decode_command	PROTO_ARGS((FILE *, Command *));
static
int	label		PROTO_ARGS((char **, char *));
static
char   *skip_white	PROTO_ARGS((char **));

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -	*/

/**
 *	compile - external function
 *	Takes a line of text and puts it into internal form.
 *
 *	Note: There is very little syntax checking on anything, although
 *	warnings are issued for unknown commands (which get OP__ZZINVALID).
 *	Note: If a label is not allowed, but one is supplied, then the
 *	label is not put into the Line.l_tag (so it can't be found).
 */
Line *
compile(text, labelok)					/* Function	*/
	char   *text;
	int     labelok;
{
	char	lab[BUFSIZ];
	Command commands[MAX_COMMANDS_PER_LINE];
	char   *bp;
	Line   *line;
	int     ncmd;

/*
printf("command: original  text='%s'\n", text);
 */
					/* Trim trailing white space	*/
	for (bp = text;  *bp;  ++bp) {	/* Skip out to end		*/

		if (*bp == ';') {	/* Truncate ';' but not ';;'	*/
			if (*(bp + 1) == ';') {
				++bp;
			} else {
				*bp-- = '\0';
			}

		} 
	}
	for (--bp;  (bp > text) && isspace(*bp);  --bp)
		;			/* Back over trailing space	*/
	*++bp = '\0';			/* Truncate after last non-white*/
	*++bp = '\0';
/*
printf("command: truncated text='%s'\n", text);
 */

	bp = text;			/* Start at the beginning	*/
/*
printf("compile:label\n");
 */
	if (label(&bp, lab)) {		/* Capture label if any		*/
		if (!labelok) {
			diag("label not allowed on line");
			strcpy(lab, "");
		}
/*
printf("  got '%s'\n", lab);
 */
	}

/*
printf("compile:commands\n");
 */
	ncmd = 0;
	while (*bp) {			/* Gather up commands		*/
/*
printf("compile: command at '%s'\n", bp);
 */
		commands[ncmd++] = command(&bp);
/*
printf("compile:command %d\n", ncmd);
decode_command(stdout, commands + ncmd - 1);
 */
	}

	line = allocLine(ncmd);
	line->l_tag  = strlen(lab) ? strdup(lab) : NULL;
	line->l_ncmd = ncmd;
/*
printf("compile:command line=%p ->l_tag='%s' ->l_ncmd=%d\n",
line, line->l_tag, line->l_ncmd);
 */
	if (ncmd) {			/* Copy commands into place	*/
		while (ncmd--) {
			line->l_cmd[ncmd] = commands[ncmd];
		}
	}
line->l_next = (Line *)NULL;
/*
decompile(stdout, line);
 */
	return line;
} /* compile */

void
decompile(f, line)					/* Function	*/
	FILE   *f;
	Line   *line;
{
	int	i;

/*
printf("decompile: line=0x%p tag='%s' ncmd=%d\n",
line, line->l_tag, line->l_ncmd);
 */
	if (line->l_tag)
		fprintf(f, "%s", line->l_tag);
	for (i = 0;  i < line->l_ncmd;  ++i)
		decode_command(f, line->l_cmd + i);
	fprintf(f, "\n");
} /* decompile */

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -	*/

/***
 *	Local Functions. (alphabetical)
 */

/**
 *	Parse out a command.
 */
static
Command
command(bpp)						/* Function	*/
	char  **bpp;
{
	Command c;
	char	diagmsg[BUFSIZ];
	char	cmd[BUFSIZ], post[BUFSIZ], arg[BUFSIZ], ch;
	int     postopts, argopts;


	c.cmd_op       = OP__NONE;
	c.cmd_postcond = NULL_MStr;
	c.cmd_arg      = NULL_MStr;

	strcpy(post, "");		/* Clear accumulators		*/
	strcpy(arg, "");

/*
printf("command: *bpp='%s'\n", *bpp);
 */
	*bpp = skip_white(bpp);		/* Get past any amount of white	*/
/*
printf("command: skip_white done... *bpp='%s'\n", *bpp);
 */
	if (**bpp == ';') {		/* Copy comment to end of line	*/
		c.cmd_op = OP__COMMENT;
		strcpy(arg, *bpp);
		*bpp += strlen(*bpp);
		goto RegularExit;
	}
					/* Gather command		*/
	*bpp = copy_non_white_upto(bpp, cmd, ":");
/*
printf("command: copy_non_white_upto done... cmd='%s' *bpp='%s'\n", cmd, *bpp);
 */

	c.cmd_op       = lkup_opcode(cmd, &postopts, &argopts);
/*
printf("command: lkup_opcode _op=%d postopts=0x%x argopts=0x%x\n",
c.cmd_op, postopts, argopts);
 */

	if (**bpp == ';') {		/* Copy comment to end of line	*/
		goto RegularExit;
	}
					/* Get postcondtional		*/
	if ((ch = **bpp) && (ch == ':')) {
/*
printf("command: postconditional\n");
 */
		(*bpp)++;		/* Skip the ':'			*/
		*bpp = copy_non_white(bpp, post);
/*
printf("command: copy_non_white post='%s' *bpp='%s'\n", post, *bpp);
 */
		if (! postopts) {
			sprintf(diagmsg,
				"post conditional \"%s\" not allowed on %s",
				post, cmd);
			diag(diagmsg);
			strcpy(post, "");
		}
	}
					/* Get argument			*/
/*
printf("command: get argument *bpp='%s'\n", *bpp);
 */
	if (**bpp == ';') {		/* Copy comment to end of line	*/
		goto RegularExit;
	}
	if ((ch = **bpp) && isspace(ch)) {
/*
printf("command: argument\n");
 */
		(*bpp)++;		/* Skip the space		*/
		*bpp = copy_non_white(bpp, arg);
/*
printf("command: copy_non_white arg='%s' *bpp='%s'\n", arg, *bpp);
 */
		if (strlen(arg)) {
					/* HACK to figure it out	*/
			switch (c.cmd_op) {
			case OP_HALT:	/* Fortunately both allow for	*/
			case OP_HANG:	/*   a postconditional		*/
				if (strlen(cmd) == 1)
					c.cmd_op = lkup_opcode(
							strcpy(cmd, "hang"),
							&postopts, &argopts);
			}
			if (argopts & ARG_NONE) {
				sprintf(diagmsg,
					"%s does not allow the argument %s",
					cmd, arg);
				diag(diagmsg);
				strcpy(arg, "");
			}
		} else {		/* No argument given		*/
			if (argopts & ARG_RQRD) {
				sprintf(diagmsg,
					"%s requires an argument", cmd);
				diag(diagmsg);
			}
		}
	}
/*
printf("command: wrapup\n");
 */

RegularExit:
	c.cmd_postcond = str2MStr(post);
	c.cmd_arg      = str2MStr(arg);
	return c;
} /* command */

/**
 *	Copy non-white space from input to a buffer.
 */
static
char *
copy_non_white(bpp, buf)				/* Function	*/
	char  **bpp;
	char   *buf;
{
/*
printf("copy_non_white:\n");
 */
	return copy_non_white_upto(bpp, buf, "");
} /* copy_non_white */

/**
 *	Copy non-white space from input to a buffer, stopping if char
 *	is in the stoplist.
 */
static
char *
copy_non_white_upto(bpp, buf, stoplist)			/* Function	*/
	char  **bpp;
	char   *buf;
	char   *stoplist;
{
	char	ch;
	char   *bp = *bpp;
	char   *cp = buf;
	int	instring = 0;

/*
printf("copy_non_white_upto(*bpp='%s', buf, stoplist='%s')\n", *bpp, stoplist);
 */

	while ((ch = *bp++)) {		/* Copy non-white space		*/
/*
printf("  got ch='%c'\n", ch);
 */
		if (ch == '"')
			instring = !instring;
		else if (!instring && (ch == ';'))
			break;
		else if (instring)
			;		/* No - op			*/
		else if (isspace(ch) || strchr(stoplist, ch))
			break;
		*cp++ = ch;
	}
	if (instring)
		diag("line terminated before closing quote");
	*cp  = '\0';			/* Terminate string in buf	*/
/*
printf("copy_non_white_upto: buf='%s' ch=0x%x ('%c') bp='%s'\n",
buf, ch, ch, ch ? bp-1 : bp);
 */
	return ch ? --bp : bp;		/* Return where we are		*/
} /* copy_non_white_upto */

/**
 *	Return the encoded (compiled) structure to source form.
 */
void
decode_command(f, cmd)					/* Function	*/
	FILE   *f;
	Command *cmd;
{
					/* Decode using long name	*/
	fprintf(f, " %s", op2name(cmd->cmd_op, 1));

	if (cmd->cmd_postcond.size)	/* Post conditional		*/
		fprintf(f, ":%.*s",
			(int)	cmd->cmd_postcond.size,
			(char *)cmd->cmd_postcond.data);
	if (cmd->cmd_op != OP__COMMENT)
		fprintf(f, " ");
	if (cmd->cmd_arg.size)		/* Argument			*/
		fprintf(f, "%.*s",
			(int)	cmd->cmd_arg.size,
			(char *)cmd->cmd_arg.data);
} /* decode_command */

/**
 *	Get a label from the input.
 *	Note: If starts with non-white, is a label, else, nothing.
 *	      This is called only when we are looking for a label.
 */
static
int
label(bpp, buf)						/* Function	*/
	char  **bpp;
	char   *buf;
{
	*bpp = copy_non_white(bpp, buf);
/*
printf("label='%s'\n", buf);
 */
	return strlen(buf);
} /* label */

/**
 *	Skip over white space in the input.
 */
static
char *
skip_white(bpp)
	char  **bpp;
{
	char   *bp = *bpp;
	char	ch;

	while ((ch = *bp++)) {		/* Skip over white space	*/
		if (!isspace(ch))
			break;
	}
	return ch ? --bp : bp;		/* Return where we are		*/
} /* skip_white */

/* End of $Source: /usr/home/ldl/langs/fm/RCS/compile.c,v $ */
